import json

import slacker
from django.contrib.auth.decorators import login_required
from django.contrib.auth.decorators import user_passes_test
from jsonview.decorators import json_view
from jsonview.exceptions import BadRequest

from orchestra.core.errors import TaskAssignmentError
from orchestra.core.errors import WorkerCertificationError
from orchestra.models import Task
from orchestra.models import Project
from orchestra.models import Worker
from orchestra.utils.revert import revert_task_to_iteration
from orchestra.utils.task_lifecycle import complete_and_skip_task
from orchestra.utils.task_lifecycle import create_subsequent_tasks
from orchestra.utils.task_lifecycle import reassign_assignment
from orchestra.utils.task_lifecycle import assign_task
from orchestra.utils.task_lifecycle import end_project

from orchestra.interface_api.project_management import project_management

import logging

logger = logging.getLogger(__name__)


def is_project_admin(user):
    return user.groups.filter(name='project_admins').exists()


@json_view
@login_required
@user_passes_test(is_project_admin)
def project_information_api(request):
    project_id = json.loads(request.body.decode())['project_id']
    try:
        return project_management.project_management_information(project_id)
    except Project.DoesNotExist:
        raise BadRequest('Project not found for the given id.')


@json_view
@login_required
@user_passes_test(is_project_admin)
def reassign_assignment_api(request):
    worker_username = json.loads(request.body.decode())['worker_username']
    try:
        worker = Worker.objects.get(user__username=worker_username)
    except Worker.DoesNotExist:
        raise BadRequest('Worker not found for the given username.')
    assignment_id = json.loads(request.body.decode())['assignment_id']

    try:
        reassign_assignment(worker.id, assignment_id)
    except (WorkerCertificationError, TaskAssignmentError) as e:
        raise BadRequest(e)


@json_view
@login_required
@user_passes_test(is_project_admin)
def revert_task_api(request):
    body = json.loads(request.body.decode())
    try:
        audit = revert_task_to_iteration(
            body['task_id'], body['iteration_id'],
            body.get('revert_before'), body.get('commit'))
    except Task.DoesNotExist as e:
        raise BadRequest(e)
    return audit


@json_view
@login_required
@user_passes_test(is_project_admin)
def complete_and_skip_task_api(request):
    task_id = json.loads(request.body.decode())['task_id']
    complete_and_skip_task(task_id)


@json_view
@login_required
@user_passes_test(is_project_admin)
def assign_task_api(request):
    worker_username = json.loads(request.body.decode())['worker_username']
    try:
        worker = Worker.objects.get(user__username=worker_username)
        task_id = json.loads(request.body.decode())['task_id']
        assign_task(worker.id, task_id)
    except (Worker.DoesNotExist,
            Task.DoesNotExist,
            WorkerCertificationError) as e:
        raise BadRequest(e)


@json_view
@login_required
@user_passes_test(is_project_admin)
def create_subsequent_tasks_api(request):
    project_id = json.loads(request.body.decode())['project_id']
    try:
        project = Project.objects.get(id=project_id)
    except Project.DoesNotExist:
        raise BadRequest('Project not found for the given id.')
    create_subsequent_tasks(project)


@json_view
@login_required
@user_passes_test(is_project_admin)
def edit_slack_membership_api(request):
    body = json.loads(request.body.decode())
    try:
        project_management.edit_slack_membership(
            body['project_id'], body['username'], body['action'])
    except slacker.Error as e:
        raise BadRequest(e)


@json_view
@login_required
@user_passes_test(is_project_admin)
def end_project_api(request):
    project_id = json.loads(request.body.decode())['project_id']
    try:
        end_project(project_id)
    except Project.DoesNotExist:
        raise BadRequest('Project not found for the given id.')



from django.db import models
from django.db.models import CASCADE
from django.utils import timezone


class DeleteMixin(object):
    """
    Overrides delete and sets `is_deleted=True` instead of deleting object.
    Intended to be used with models that have an `is_deleted` field.
    """
    def delete(self, *args, **kwargs):
        # Implement cascading deletes.
        # NOTE(lydia): This ignores any other constraints passed to on_delete.
        fields = self._meta.get_fields(include_hidden=True)
        for field in fields:
            # Pass on fields that are not many-to-one relationships.
            if not ((field.one_to_many or field.one_to_one) and
                    field.auto_created):
                continue

            # Check if foreign key is set up with on_delete=CASCADE.
            if field.on_delete != CASCADE:
                continue

            # Query for all related objects and delete them.
            query_args = {field.field.name: self.id}
            related_objs = field.related_model.objects.filter(
                **query_args)
            for obj in related_objs:
                obj.delete()

        # Mark current object as deleted.
        self.is_deleted = True
        self.save()


class BaseModelManager(models.Manager):
    """
    Model manager intended to be used with models with an `is_deleted` field.
    Overrides the initial QuerySet to filter for objects where `is_deleted`
    is false.
    """
    def get_queryset(self):
        return (super(BaseModelManager, self).get_queryset()
                .filter(is_deleted=False))


class BaseModel(DeleteMixin, models.Model):
    """
    Abstract base class models which defines created_at and is_deleted fields.

    Attributes:
        created_at (datetime.datetime):
            Datetime at which the model is created.
        is_deleted (boolean):
            If value is True, mdoel is deleted. Default is False.
    """
    objects = BaseModelManager()

    created_at = models.DateTimeField(default=timezone.now)
    is_deleted = models.BooleanField(default=False)

    class Meta:
        abstract = True



import django_filters

from rest_framework import filters

from orchestra.models import TimeEntry


class TimeEntryFilter(filters.FilterSet):
    min_date = django_filters.DateFilter(name="date", lookup_type='gte')
    max_date = django_filters.DateFilter(name="date", lookup_type='lte')

    class Meta:
        model = TimeEntry
        fields = ['min_date', 'max_date']



from django.conf.urls import include
from django.conf.urls import patterns
from django.conf.urls import url

from orchestra.views import index
from orchestra.views import OrchestraRegistrationView
from orchestra.views import status


urlpatterns = patterns(
    '',
    url(r'^api/',
        include('orchestra.api_urls', namespace='orchestra')),
    url(r'^app/?', index, name='index'),
    url(r'^status/', status, name='status'),

    # We have to override the register endpoint to emit the
    # `orchestra_user_registered` signal
    url(r'^accounts/register/$',
        OrchestraRegistrationView.as_view(),
        name='registration_register'),
)



import datetime

from dateutil.parser import parse

from orchestra.core.errors import ModelSaveError
from orchestra.models import PayRate
from orchestra.models import Worker
from orchestra.models import WorkerCertification
from orchestra.signals import orchestra_user_registered
from orchestra.tests.helpers import OrchestraTestCase
from orchestra.tests.helpers.fixtures import CertificationFactory
from orchestra.tests.helpers.fixtures import WorkerCertificationFactory
from orchestra.tests.helpers.fixtures import UserFactory
from orchestra.tests.helpers.fixtures import setup_models


class ModelsTestCase(OrchestraTestCase):

    def setUp(self):
        super().setUp()
        setup_models(self)
        self.workers[0] = Worker.objects.get(user__username='test_user_1')
        self.worker = self.workers[0]

    def test_certification_roles(self):
        """ Ensure that workers can be certified at multiple roles. """
        certification = CertificationFactory(
            slug='cat_herding', name='Cat herding',
            workflow=self.workflows['w1'])
        WorkerCertificationFactory(
            certification=certification,
            worker=self.workers[0],
            role=WorkerCertification.Role.ENTRY_LEVEL)
        WorkerCertificationFactory(
            certification=certification,
            worker=self.workers[0],
            role=WorkerCertification.Role.REVIEWER)

    def test_reviewer_requires_entry_level(self):
        """ Workers must have entry-level certification before they review. """
        certification = CertificationFactory(
            slug='cat_herding', name='Cat herding',
            workflow=self.workflows['w1'])
        with self.assertRaises(ModelSaveError):
            WorkerCertificationFactory(
                certification=certification,
                worker=self.workers[0],
                role=WorkerCertification.Role.REVIEWER)

    def test_worker_autocreation(self):
        """ When new users register, worker objects should be autocreated. """

        # Create a user
        user = UserFactory(username='test_registration_user',
                           password='test',
                           email='test_registration_user@test.com')

        # There should be no worker yet
        self.assertFalse(Worker.objects.filter(user=user).exists())

        # Fake registering the user
        orchestra_user_registered.send(
            sender=self.__class__, user=user, request=None)

        # Expect the worker object to be created
        self.assertTrue(Worker.objects.filter(user=user).exists(),
                        'Worker not autocreated on User registration')

    def test_pay_rate_creation(self):
        """
        Ensure that PayRates are created with valid ranges and do not
        overlap with existing entries.
        """
        start_stop_dates = [
            (('2016-01-01', '2016-01-03'), ('2016-01-02', None), False),
            (('2016-01-01', None), ('2016-02-01', None), False),
            (('2016-01-01', None), ('2015-12-01', '2016-01-01'), False),
            (('2016-01-01', '2016-01-10'), ('2016-01-03', '2016-01-08'),
             False),
            (('2016-01-01', '2016-01-10'), ('2015-12-30', '2016-01-03'),
             False),
            (('2016-01-01', '2016-01-10'), ('2016-01-03', '2016-01-18'),
             False),
            (('2016-01-01', '2016-01-10'), ('2015-12-30', '2016-01-18'),
             False),
            (('2016-01-01', '2016-01-10'), ('2016-01-30', None), True),
            (('2016-01-01', '2016-01-10'), ('2016-01-30', '2016-02-18'), True),
            (('2016-03-02', None), ('2016-01-01', '2016-03-01'), True)
        ]

        def _parse_date(date):
            if date:
                start_date = parse(date).date()
            else:
                start_date = None
            return start_date

        def _verify_failure(saved, new, success):
            start_date_saved = _parse_date(saved[0])
            end_date_saved = _parse_date(saved[1])
            PayRate.objects.create(worker=self.worker,
                                   hourly_rate=10,
                                   hourly_multiplier=1.2,
                                   start_date=start_date_saved,
                                   end_date=end_date_saved)
            start_date_new = _parse_date(new[0])
            end_date_new = _parse_date(new[1])
            payrate = PayRate(worker=self.worker,
                              hourly_rate=10,
                              hourly_multiplier=1.2,
                              start_date=start_date_new,
                              end_date=end_date_new)
            if success:
                payrate.save()
            else:
                with self.assertRaises(ModelSaveError):
                    payrate.save()
            PayRate.objects.all().delete()

        for saved, new, success in start_stop_dates:
            _verify_failure(saved, new, success)

    def test_pay_rate_update(self):
        """ Verify that we don't detect overlap collision with itself. """
        payrate = PayRate.objects.create(
            worker=self.worker,
            hourly_rate=10,
            hourly_multiplier=1.2,
            start_date=datetime.date(2016, 1, 1),
            end_date=None)
        payrate.end_date = datetime.date(2016, 3, 1)
        payrate.save()
